      program main_bc_wrfchem

!<DESCRIPTION>
!
! Program wrfchembc.
!
! Authors:  Rainer Schmitz (University of Chile - Santiago, Chile)
!           Steven Peckham (NOAA/ESRL/GSD - Boulder, CO)
!
! Main program of WRF/Chem global boundary condition generation code.  Responsible for 
! starting up the model, reading in configuration data, defining and initializing
! the top-level domain, either from initial or restart data, setting up time-keeping, and
! then calling the proper routine to get the global boundary condition data.
! After the boundary condition data construction is completed and the data is added to the
! wrfbdy_d01 data file, the model is properly shut down.
!
! Other files needed:
!  module_mozart_lib.f90  
!  module_raqms_lib.f90  
!  module_wrfchem_lib.f90
!  wrfchembc_namelist.input
!  Makefile
!
! The namelist input data file for the program is structured like the following
! &control
! 
! dir_wrf = '/data/wrfchem_data/wrfchem_raqmsbc/'   ! Run directory, wrf data directory
! fnb_wrf  = 'wrfbdy_d01'                           ! WRF boundary condition data file
! fni_wrf  = 'wrfinput_d01'                         ! WRF initial condition data file

! chem_bc_opt = 2                                   ! Global Model data used: 1=MOZART, 2=RAQMS
! dir_global = '/data/RAQMS_DATA/'                  ! Global model data directory
! fn_global  = 'uwhyb_09_02_2006_00Z.chem.assim.nc' ! Global model data file name
! 
! nspec = 10                                        ! number of species listed by name in specname
! 
! specname = 'o3',                                  ! name of chemical species to be added to boundary conditions
!            'co',
!            'no',
!            'no2',
!            'n2o5',
!            'h2o2',
!            'ho',
!            'ho2',
!            'eth',
!            'iso'
! /
!
! To compile, type "make". The make file is set-up to compile the code for a 
! single processor using the Intel compiler (We assume you are using a 
! linux cluster).  The Makefile will need to be modified to work with other 
! computer systems.
!
! To execute the program, type 
!
!  wrfchembc < wrfchembc_namelist.input
!
!</DESCRIPTION>

      use module_wrfchem_lib
      use module_mozart_lib
      use module_raqms_lib

      implicit none

!     parameters
      integer, parameter :: maxsize = 150
      integer, parameter :: kin = 15

!     control variables
      character(len=maxsize) :: dir_global
      character(len=maxsize) :: fnb_wrf, fni_wrf, dir_wrf
      integer :: nspec, chem_bc_opt
      character(len=12),DIMENSION(500) :: specname
!tks
      logical :: do_bc = .true.
      logical :: do_ic = .false.
      namelist /control/ dir_global, fnb_wrf, fni_wrf, dir_wrf, chem_bc_opt,  &
                         fn_global, nspec, specname, do_bc, do_ic

!tks
!     wrf variables 
      integer :: iptop_wrf
      real    :: ptop_wrf
      real, allocatable :: znu(:)
      real, allocatable :: xlon(:,:), xlat(:,:)

!     global file name
      character(len=maxsize) :: fn_global

      real, allocatable :: spc_ic(:,:,:)
!     species vmr
      real, allocatable :: bcxs(:,:,:)
      real, allocatable :: bcxe(:,:,:)
      real, allocatable :: bcys(:,:,:)
      real, allocatable :: bcye(:,:,:)
!     species vmr tendencies
      real, allocatable :: btxs(:,:,:)
      real, allocatable :: btxe(:,:,:)
      real, allocatable :: btys(:,:,:)
      real, allocatable :: btye(:,:,:)
!     species vmr old
      real, allocatable :: bcxso(:,:,:)
      real, allocatable :: bcxeo(:,:,:)
      real, allocatable :: bcyso(:,:,:)
      real, allocatable :: bcyeo(:,:,:)
!     species vmr temp1
      real, allocatable :: t1xs(:,:,:,:)
      real, allocatable :: t1xe(:,:,:,:)
      real, allocatable :: t1ys(:,:,:,:)
      real, allocatable :: t1ye(:,:,:,:)
!     species vmr temp2
      real, allocatable :: t2xs(:,:,:,:)
      real, allocatable :: t2xe(:,:,:,:)
      real, allocatable :: t2ys(:,:,:,:)
      real, allocatable :: t2ye(:,:,:,:)
!     other working variables
      integer            :: ns, ios, it, i_con
      character(len=maxsize) :: fnb, fni
      character(len=12)      :: spec1
      logical            :: time
      real               :: t_ratio, it_ratio, f_int

      print *,"*********************************************************************"
      print *,"*   PROGRAM: WRF/CHEM GLOBAL BOUNDARIES                             *"
      print *,"*              FOR WRF/CHEM VERSION 2.2                             *"
      print *,"*                                                                   *"
      print *,"*    PLEASE REPORT ANY BUGS TO WRF/CHEM HELP at                     *"
      print *,"*                                                                   *"
      print *,"*              wrfchemhelp.gsd@noaa.gov                             *"
      print *,"*                                                                   *"
      print *,"*********************************************************************"

!     read control variables
      read( 5, nml=control )

!     intialize module_wrfchem_lib
      fnb = trim(dir_wrf)//adjustl(fnb_wrf)
      fni = trim(dir_wrf)//adjustl(fni_wrf)
      print*,'dir_wrf ',trim(dir_wrf)


!tks      call init_wrfchem_lib( fnb, fni, specname, nspec, ntime )
      call init_wrfchem_lib( fnb, fni, specname, nspec, ntime, do_bc, do_ic )

       print*,'tks after init_wrfchem' 



!     read top ref. pressure from wrf input data

      call wrfchem_readinteger( 'P_TOP', iptop_wrf )
      write(6,*)'ptop A ',iptop_wrf
      call flush(6)
      ptop_wrf = float( iptop_wrf )*.01        ! convert to mb


!     read eta values on half (mass) levels from wrfchem
      allocate( znu(nz) )
      call wrfchem_read2d( 'ZNU', znu )
      print*, 'read eta values on half (mass) levels'

!     read longitudes and latitudes from wrfchem input data
      allocate( xlon(nx,ny), xlat(nx,ny) )
      call wrfchem_read3d( 'XLONG', xlon )
      call wrfchem_read3d( 'XLAT',  xlat )
      where( xlon < 0.0 )
        xlon = xlon + 360.0
      endwhere
      print*,'read longitudes and latitudes from wrfchem input data'

!     initializie global data module
      fni = trim(dir_global)//adjustl(fn_global)
      if(chem_bc_opt == 1) then
        call init_mozart_lib( fni, ptop_wrf, xlon, xlat, znu, nx, ny, nz )
      elseif(chem_bc_opt == 2) then
        call init_raqms_lib( fni, ptop_wrf, xlon, xlat, znu, nx, ny, nz )
      else
        print*,'Error: Improper chemical boundary data specified, set chem_bc_opt in namelist'
        stop 999
      endif



      print*,'FNI 1 ',fni

!     read global boundary data, interpolte and then write as IC
!tks
      if ( do_ic ) then
         allocate( spc_ic(nx,ny,nz) )
      end if
!tks
      allocate( bcxs(ny,nz,nw) )
      allocate( bcxe(ny,nz,nw) )
      allocate( bcys(nx,nz,nw) )
      allocate( bcye(nx,nz,nw) )
      allocate( bcxso(ny,nz,nw) )
      allocate( bcxeo(ny,nz,nw) )
      allocate( bcyso(nx,nz,nw) )
      allocate( bcyeo(nx,nz,nw) )
      allocate( btxs(ny,nz,nw) )
      allocate( btxe(ny,nz,nw) )
      allocate( btys(nx,nz,nw) )
      allocate( btye(nx,nz,nw) )
      allocate( t1xs(ny,nz,nw,nspec) )
      allocate( t1xe(ny,nz,nw,nspec) )
      allocate( t1ys(nx,nz,nw,nspec) )
      allocate( t1ye(nx,nz,nw,nspec) )
      allocate( t2xs(ny,nz,nw,nspec) )
      allocate( t2xe(ny,nz,nw,nspec) )
      allocate( t2ys(nx,nz,nw,nspec) )
      allocate( t2ye(nx,nz,nw,nspec) )

!---------------------------------------------------------------------------------------------------
!     time check of global data and WRF data
!---------------------------------------------------------------------------------------------------
      if(chem_bc_opt == 1) then
         if ( year_start_m == year_start .and.  month_start_m == month_start .and. &
              day_start_m == day_start   .and.  hour_start_m == hour_start   .and. &
              year_end_m >= year_end     .and.  month_end_m >= month_end     .and. &
              day_end_m >= day_end       .and.  hour_end_m >= hour_end       ) then
            time = .true.
         else
            print*, 'No data in MOZART BC file for this WRF run but we proceed anyway'
            time = .true.
        !    stop
         end if
      else
         if ( year_start_r == year_start .and.  month_start_r == month_start .and. &
              day_start_r == day_start   .and.  hour_start_r == hour_start   .and. &
              year_end_r >= year_end     .and.  month_end_r >= month_end     .and. &
              day_end_r >= day_end       .and.  hour_end_r >= hour_end       ) then
            time = .true.
         else
            print*, 'No data in RAQMS BC file for this WRF run but we proceed anyway'
            time = .true.
         end if
      end if
      
      it_ratio = 1.
      if ( time .and. chem_bc_opt == 1) then
         t_ratio  = float(dt)/float(dt_m)
         print*, 'Times are ok.'
         if ( t_ratio /= 1. ) then
            it_ratio = float(dt_m)/float(dt)
            t_ratio  = float(dt)/float(dt_m)
          endif
      else if ( time .and. chem_bc_opt == 2) then
         t_ratio  = float(dt)/float(dt_r)
         print *, 'Times are ok.' 
         if ( t_ratio /= 1. ) then
            it_ratio = float(dt_r)/float(dt)
            t_ratio  = float(dt)/float(dt_r)
          endif
      end if
      print *, 'We have to interpolate in time with factor: ', t_ratio
      
      i_con = 1
      f_int = 1.
      if ( chem_bc_opt == 1) then
      do ns = 1, nspec
         call mozart_interpolate4d( specname(ns), t1xs(:,:,:,ns), t1xe(:,:,:,ns), t1ys(:,:,:,ns), &
                           t1ye(:,:,:,ns), nx, ny, nz, nw, i_con )
      end do
      else
      do ns = 1, nspec
!tks  write RAQMS initial condtions if desired
            if ( do_ic ) then !! initial cond.
               if ( chem_bc_opt == 2 ) then !! RAQMS
                  if(ns.eq.1)print*,'WRITE OUT RAQMS IC'
                  call raqms_interpolate4d_ic( specname(ns), spc_ic, nx, ny, nz, ptop_wrf, znu )
                  call wrfchem_write4d_ic( specname(ns), spc_ic )
               end if
            end if
        if( do_bc) then
!tks
         call raqms_interpolate4d( specname(ns), t1xs(:,:,:,ns), t1xe(:,:,:,ns), t1ys(:,:,:,ns), &
                           t1ye(:,:,:,ns), nx, ny, nz, nw, i_con,znu,ptop_wrf )   !tks added znu and ptop_wrf
        endif
      end do
      endif

      if( do_bc)then
      i_con = 2
!     do it = 1, ntime  + 1
      do it = 1, ntime 
      
         do ns = 1, nspec

            if ( mod(it-1,int(it_ratio)) == 0. ) then
               t2xs(:,:,:,ns) = t1xs(:,:,:,ns)
               t2xe(:,:,:,ns) = t1xe(:,:,:,ns)
               t2ys(:,:,:,ns) = t1ys(:,:,:,ns)
               t2ye(:,:,:,ns) = t1ye(:,:,:,ns)
      if ( chem_bc_opt == 1) then
               call mozart_interpolate4d( specname(ns), t1xs(:,:,:,ns), t1xe(:,:,:,ns), t1ys(:,:,:,ns), &
                                 t1ye(:,:,:,ns), nx, ny, nz, nw, i_con )
      else
               if(ns == 1 ) call re_init_raqms( fni, it, dt_r )
               call raqms_interpolate4d( specname(ns), t1xs(:,:,:,ns), t1xe(:,:,:,ns), t1ys(:,:,:,ns), &
                                 t1ye(:,:,:,ns), nx, ny, nz, nw, i_con , znu, ptop_wrf)    !tks added znu and ptop_wrf
      endif
            endif
!tks           if (it >=2) then
!tks              bcxso = bcxs
!tks              bcxeo = bcxe
!tks              bcyso = bcys
!tks              bcyeo = bcye

               bcxso = t2xs(:,:,:,ns)      !tks
               bcxeo = t2xe(:,:,:,ns)      !tks
               bcyso = t2ys(:,:,:,ns)      !tks
               bcyeo = t2ye(:,:,:,ns)      !tks
!tks            end if
            ! BC

!tks  want bcxs to be t1xs, need f_int=0 in that case, not f_int=1

            f_int=0.       !tks
            bcxs = f_int * t2xs(:,:,:,ns) + ( 1. - f_int ) * t1xs(:,:,:,ns)
            bcxe = f_int * t2xe(:,:,:,ns) + ( 1. - f_int ) * t1xe(:,:,:,ns)
            bcys = f_int * t2ys(:,:,:,ns) + ( 1. - f_int ) * t1ys(:,:,:,ns)
            bcye = f_int * t2ye(:,:,:,ns) + ( 1. - f_int ) * t1ye(:,:,:,ns)

!tks            if (it >=2) then
!tks compute tendency for it - (it-1)
               ! Tendencies
               btxs = ( bcxs - bcxso )/(dt*3600.)
               btxe = ( bcxe - bcxeo )/(dt*3600.)
               btys = ( bcys - bcyso )/(dt*3600.)
               btye = ( bcye - bcyeo )/(dt*3600.)
!tks            end if
!          if( ns == 1) then
!            print *,' TIMESTEP: ',it
!            print *,' '
!            print *,'f_int ',1. - f_int, f_int
!            print *,' '
!            print *,t1xs(1,:,1,ns)
!            print *,' '
!            print *,t2xs(1,:,1,ns)
!            print *,' '
!            print *,bcxs(1,:,1)
!            print *,' '
!          endif
            
!tks bcxso holds value at t0, bcxs holds value at t, tendency btxs=(bcsx-bcsxo)/deltat   ie, (t-t0)/deltat
!In Wrf/chem module_input_chem_data.F, CALL bdy_chem_value_gcm ( chem(i,k,j),chem_bys(i,k,1),chem_btys(i,k,1),dt,ic)
!SUBROUTINE bdy_chem_value_gcm ( chem, chem_b, chem_bt, dt,ic)
!chem=max(epsilc,chem_b + chem_bt * dt)
!where chem_b is bdy value at t0, chem_bt is tendency (t1-t0)/deltat, dt=model timestep


!tks expects the data

!tks           call wrfchem_write4d( specname(ns), bcxs, bcxe, bcys, bcye, btxs, btxe, btys, btye, it )
!           print*,'max,min of west bdy ',specname(ns),maxval(bcxso),minval(bcxso)

            call wrfchem_write4d_bc( specname(ns), bcxso, bcxeo, bcyso, bcyeo, btxs, btxe, btys, btye, it )

            if(ns.eq.14)print*,'tks,ns,it ',ns,it,bcxs(10,10,1),bcxso(10,10,1),btxs(10,10,1)

         enddo
         f_int = f_int - t_ratio
         if ( mod(it,int(it_ratio)) == 0. ) then
            i_con = i_con + 1
            f_int = 1.
            if(t_ratio < .5) f_int = f_int - t_ratio
         end if
      end do

      endif    !tks end if (do_bc)
!     exit from libs
      call exit_wrfchem_lib( )
      if(chem_bc_opt == 1) then
        call exit_mozart_lib( )
      elseif(chem_bc_opt == 2) then
        call exit_raqms_lib( )
      endif

!     deallocate memory space
      deallocate( znu, xlon, xlat, bcxs, bcxe, bcys, bcye,    &
           bcxso, bcxeo, bcyso, bcyeo, btxs, btxe, btys, btye,&
           t1xs, t1xe, t1ys, t1ye, t2xs, t2xe, t2ys, t2ye  )

      if ( do_ic ) then
         deallocate( spc_ic )
      end if
      print*,'bc_wrfchem completed successfully'
    end program main_bc_wrfchem
