
      module module_mozart_lib

      implicit none

!     include files
      include 'netcdf.inc'

!     public readprocedures
      public  :: init_mozart_lib, mozart_interpolate4d, exit_mozart_lib

!     private procedures
      private :: handle_error_moz, translate_radm2_mozart

!     public variables
      integer, public :: nlon, nlat, nlev                    ! data dimension length
      integer, public  :: year_start_m, month_start_m, day_start_m
      integer, public  :: hour_start_m, minute_start_m, second_start_m
      integer, public  :: year_end_m, month_end_m, day_end_m
      integer, public  :: hour_end_m
      integer, public  :: dt_m, ntime_m

!     variables for reading netCDF data
      integer, private :: ncid                               ! netCDF file ID
      integer, private :: nstt(4), ncnt(4)                   ! start index and counter array
      integer, private :: start_index = 1                    ! monthly mean data
      integer, private  :: ntime      ! data dimension length
      integer, private :: nscalar, nstring           ! data dimension length
      integer, private, allocatable :: date(:), datesec(:) 

!     variables used by interpolation
      integer, private, allocatable :: ix(:,:,:), jy(:,:,:), kz(:,:,:,:)  ! index used by interpolation
      real, private, allocatable :: ax(:,:,:), by(:,:,:), cz(:,:,:,:)           ! weight coef. all domain
      real, private, allocatable :: axxs(:,:,:), byxs(:,:,:), czxs(:,:,:,:)     ! weight coef. west
      real, private, allocatable :: axxe(:,:,:), byxe(:,:,:), czxe(:,:,:,:)     ! weight coef. east
      real, private, allocatable :: axys(:,:,:), byys(:,:,:), czys(:,:,:,:)     ! weight coef. south
      real, private, allocatable :: axye(:,:,:), byye(:,:,:), czye(:,:,:,:)     ! weight coef. north

      real, private, allocatable :: mozval(:,:,:), tmpval(:,:,:), mozval2(:,:,:,:)

      contains

!---------------------------------------------------------------------------------------------------
!     initialize netcdf file
!---------------------------------------------------------------------------------------------------

      subroutine init_mozart_lib( mozart_fn, ptop_wrf, x2d, y2d, znu, nx, ny, nz )

      implicit none

!     input arguments
      integer, intent(in)      :: nx, ny, nz
      character(*), intent(in) :: mozart_fn 
      real, intent(in)         :: ptop_wrf
      real, intent(in)         :: x2d(nx,ny), y2d(nx,ny)
      real, intent(in)         :: znu(nz)

!     local argumets
      integer :: status
      integer :: dimid, varid
      integer :: i, j, k, n 
      integer :: vid
      real, allocatable :: x1d(:), y1d(:)
      real, allocatable :: hyam(:), hybm(:)
      real, allocatable :: p_moz(:), p_wrf(:)
      real, allocatable :: ps_moz(:,:), ps_wrf(:,:)
      character(100)    :: dtstrings, dtstringe, dtstringf

!     open the input netCDF file
      status = nf_open( mozart_fn, nf_nowrite, ncid )
      if( status /= nf_noerr ) call handle_error_moz( status )

!     get the longitude dimension length of the MOZART data
      status = nf_inq_dimid( ncid, 'lon', dimid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_inq_dimlen( ncid, dimid, nlon )
      if( status /= nf_noerr )  call handle_error_moz( status )

!     get the latitude dimension length of the MOZART data
      status = nf_inq_dimid( ncid, 'lat', dimid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_inq_dimlen( ncid, dimid, nlat )
      if( status /= nf_noerr )  call handle_error_moz( status )

!     get the vertical dimension length of the MOZART data
      status = nf_inq_dimid( ncid, 'lev', dimid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_inq_dimlen( ncid, dimid, nlev )
      if( status /= nf_noerr )  call handle_error_moz( status )

!     read sigma levels
      allocate( hyam(nlev), hybm(nlev) )

      status = nf_inq_varid( ncid, 'hyam', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_get_var_real( ncid, varid, hyam )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_inq_varid( ncid, 'hybm', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_get_var_real( ncid, varid, hybm )
      if( status /= nf_noerr )  call handle_error_moz( status )

!     read longitudes
      allocate( x1d(nlon), y1d(nlat) )

      status = nf_inq_varid( ncid, 'lon', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_get_var_real( ncid, varid, x1d )
      if( status /= nf_noerr )  call handle_error_moz( status )

!     read latitudes
      status = nf_inq_varid( ncid, 'lat', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_get_var_real( ncid, varid, y1d )
      if( status /= nf_noerr )  call handle_error_moz( status )

!     read start and end date
      status = nf_inq_dimid( ncid, 'time', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_inq_dimlen( ncid, varid, nstring )
      if( status /= nf_noerr ) call handle_error_moz( status )

      ntime_m = nstring

      allocate ( date(nstring) )
      allocate (datesec(nstring) )

      status = nf_inq_varid( ncid, 'date', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_get_var_int( ncid, varid, date )

      status = nf_inq_varid( ncid, 'datesec', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      status = nf_get_var_int( ncid, varid, datesec )

      year_start_m   = floor(float(date(1)/10000))
      month_start_m  = floor(float(date(1)/100-year_start_m*100))
      day_start_m    = float(date(1)-year_start_m*10000-month_start_m*100)
      hour_start_m   = datesec(1)/3600
      
      year_end_m   = floor(float(date(nstring)/10000))
      month_end_m  = floor(float(date(nstring)/100-year_end_m*100))
      day_end_m    = float(date(nstring)-year_end_m*10000-month_end_m*100)
      hour_end_m   = datesec(nstring)/3600

      ! temporal resolution of mozart file
      dt_m         = abs(datesec(2)/3600 - datesec(1)/3600)

!     allocate memory space to store interpolation coef.
      allocate( ax(0:1,nx,ny), by(0:1,nx,ny), cz(0:1,nx,ny,nz) )
      allocate( axxs(0:1,nx,ny), byxs(0:1,nx,ny), czxs(0:1,nx,ny,nz) )
      allocate( axxe(0:1,nx,ny), byxe(0:1,nx,ny), czxe(0:1,nx,ny,nz) )
      allocate( axys(0:1,nx,ny), byys(0:1,nx,ny), czys(0:1,nx,ny,nz) )
      allocate( axye(0:1,nx,ny), byye(0:1,nx,ny), czye(0:1,nx,ny,nz) )
      allocate( ix(0:1,nx,ny), jy(0:1,nx,ny), kz(0:1,nx,ny,nz) )

!     horizontal interpolation coefs.

      ! all domain
      do i = 1, nx
      do j = 1, ny 
         do n = 1, nlon
            if( x2d(i,j) < x1d(n) ) exit
         enddo
         
         ix(0,i,j) = min( nlon-1, max(n-1,1) )
         ix(1,i,j) = ix(0,i,j) + 1
         ax(0,i,j) = ( x2d(i,j) - x1d(ix(0,i,j)) )/( x1d(ix(1,i,j)) - x1d(ix(0,i,j)) )
         ax(1,i,j) = 1.0 - ax(0,i,j)
         
         do n = 1, nlat
            if( y2d(i,j) < y1d(n) ) exit
         enddo
         
         jy(0,i,j) = min( nlat-1, max(n-1,1) )
         jy(1,i,j) = jy(0,i,j) + 1
         by(0,i,j) = ( y2d(i,j) - y1d(jy(0,i,j)) )/( y1d(jy(1,i,j)) - y1d(jy(0,i,j)) )
         by(1,i,j) = 1.0 - by(0,i,j)
         
      enddo
   end do

      ! west
      i = 1
      do j = 1, ny 
         do n = 1, nlon
            if( x2d(i,j) < x1d(n) ) exit
         enddo
         
         axxs(0,i,j) = ( x2d(i,j) - x1d(ix(0,i,j)) )/( x1d(ix(1,i,j)) - x1d(ix(0,i,j)) )
         axxs(1,i,j) = 1.0 - axxs(0,i,j)
         
         do n = 1, nlat
            if( y2d(i,j) < y1d(n) ) exit
         enddo
         
         byxs(0,i,j) = ( y2d(i,j) - y1d(jy(0,i,j)) )/( y1d(jy(1,i,j)) - y1d(jy(0,i,j)) )
         byxs(1,i,j) = 1.0 - byxs(0,i,j)
         
      enddo

      ! east
      i = nx
      do j = 1, ny 
         
         do n = 1, nlon
            if( x2d(i,j) < x1d(n) ) exit
         enddo
         
         axxe(0,i,j) = ( x2d(i,j) - x1d(ix(0,i,j)) )/( x1d(ix(1,i,j)) - x1d(ix(0,i,j)) )
         axxe(1,i,j) = 1.0 - axxe(0,i,j)
         
         do n = 1, nlat
            if( y2d(i,j) < y1d(n) ) exit
         enddo
         
         byxe(0,i,j) = ( y2d(i,j) - y1d(jy(0,i,j)) )/( y1d(jy(1,i,j)) - y1d(jy(0,i,j)) )
         byxe(1,i,j) = 1.0 - byxe(0,i,j)
         
      enddo

      ! north
      j = ny
      do i = 1, nx
         
         do n = 1, nlon
            if( x2d(i,j) < x1d(n) ) exit
         enddo
         
         axye(0,i,j) = ( x2d(i,j) - x1d(ix(0,i,j)) )/( x1d(ix(1,i,j)) - x1d(ix(0,i,j)) )
         axye(1,i,j) = 1.0 - axye(0,i,j)
         
         do n = 1, nlat
            if( y2d(i,j) < y1d(n) ) exit
         enddo
         
         byye(0,i,j) = ( y2d(i,j) - y1d(jy(0,i,j)) )/( y1d(jy(1,i,j)) - y1d(jy(0,i,j)) )
         byye(1,i,j) = 1.0 - byye(0,i,j)
         
      enddo
 
      ! south
      j = 1 
      do i = 1, nx

        do n = 1, nlon
          if( x2d(i,j) < x1d(n) ) exit
        enddo

        axys(0,i,j) = ( x2d(i,j) - x1d(ix(0,i,j)) )/( x1d(ix(1,i,j)) - x1d(ix(0,i,j)) )
        axys(1,i,j) = 1.0 - axys(0,i,j)

        do n = 1, nlat
          if( y2d(i,j) < y1d(n) ) exit
        enddo

        byys(0,i,j) = ( y2d(i,j) - y1d(jy(0,i,j)) )/( y1d(jy(1,i,j)) - y1d(jy(0,i,j)) )
        byys(1,i,j) = 1.0 - byys(0,i,j)

      enddo

!     read surface pressure and interpolate ...
      allocate( ps_moz(nlon,nlat), ps_wrf(nx,ny) )

      status = nf_inq_varid( ncid, 'PS', varid )
      if( status /= nf_noerr )  call handle_error_moz( status )

      nstt(1:3) = (/ 1, 1, start_index /)
      ncnt(1:3) = (/ nlon, nlat, 1 /)
      status = nf_get_vara_real( ncid, varid, nstt(1:3), ncnt(1:3), ps_moz )
      if( status /= nf_noerr ) call handle_error_moz( status )

      do j = 1, ny 
      do i = 1, nx 
        ps_wrf(i,j) = ps_moz(ix(0,i,j),jy(0,i,j))*ax(1,i,j)*by(1,i,j) +   &
                      ps_moz(ix(0,i,j),jy(1,i,j))*ax(1,i,j)*by(0,i,j) +   &
                      ps_moz(ix(1,i,j),jy(0,i,j))*ax(0,i,j)*by(1,i,j) +   &
                      ps_moz(ix(1,i,j),jy(1,i,j))*ax(0,i,j)*by(0,i,j)
      enddo
      enddo


!     vertical interpolation coefs.
      allocate( p_moz(nlev), p_wrf(nz) )
      
      !west
      i = 1
      do j = 1, ny
         
         p_moz(:) = log( hybm(:)*ps_wrf(i,j) + hyam(:)*1.0e5 )
         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

         
         do k = 1, nz

            do n = 1, nlev 
               if( p_wrf(k) < p_moz(n) ) exit
            enddo
            
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            czxs(0,i,j,k) = ( p_wrf(k) - p_moz(kz(0,i,j,k)) )/( p_moz(kz(1,i,j,k)) - p_moz(kz(0,i,j,k)) )
            czxs(1,i,j,k) = 1.0 - czxs(0,i,j,k)
            
         enddo

      end do

      ! east
      i = nx
      do j = 1, ny
         
         p_moz(:) = log( hybm(:)*ps_wrf(i,j) + hyam(:)*1.0e5 )
         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )
         
         do k = 1, nz
            
            do n = 1, nlev 
               if( p_wrf(k) < p_moz(n) ) exit
            enddo
            
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            czxe(0,i,j,k) = ( p_wrf(k) - p_moz(kz(0,i,j,k)) )/( p_moz(kz(1,i,j,k)) - p_moz(kz(0,i,j,k)) )
            czxe(1,i,j,k) = 1.0 - czxe(0,i,j,k)
            
         enddo
      enddo

      ! north
      j = ny
      do i = 1, nx
         
         p_moz(:) = log( hybm(:)*ps_wrf(i,j) + hyam(:)*1.0e5 )
         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )
         
         do k = 1, nz
            
            do n = 1, nlev 
               if( p_wrf(k) < p_moz(n) ) exit
            enddo
            
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            czye(0,i,j,k) = ( p_wrf(k) - p_moz(kz(0,i,j,k)) )/( p_moz(kz(1,i,j,k)) - p_moz(kz(0,i,j,k)) )
            czye(1,i,j,k) = 1.0 - czye(0,i,j,k)
            
         enddo
         
      enddo
      
      ! south
      j = 1
      do i = 1, nx

        p_moz(:) = log( hybm(:)*ps_wrf(i,j) + hyam(:)*1.0e5 )
        p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

        do k = 1, nz

          do n = 1, nlev 
            if( p_wrf(k) < p_moz(n) ) exit
          enddo

          kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
          kz(1,i,j,k) = kz(0,i,j,k) + 1
          czys(0,i,j,k) = ( p_wrf(k) - p_moz(kz(0,i,j,k)) )/( p_moz(kz(1,i,j,k)) - p_moz(kz(0,i,j,k)) )
          if( czys(0,i,j,k) > 1.)  czys(0,i,j,k) = 1.
          czys(1,i,j,k) = 1.0 - czys(0,i,j,k)

        enddo

      enddo

!     release memory space not used
      deallocate( x1d, y1d, hyam, hybm, ps_moz, ps_wrf, p_moz, p_wrf )

!     allocate memory space for reading
      allocate( mozval(nlon,nlat,nlev), tmpval(nlon,nlat,nlev), mozval2(nlon,nlat,nlev,2) )

      end subroutine init_mozart_lib

!---------------------------------------------------------------------------------------------------
!     interpolate four-dimensional field ... 
!---------------------------------------------------------------------------------------------------

      subroutine mozart_interpolate4d( wrfspn, wrfxs, wrfxe, wrfys, wrfye, nx, ny, nz, nw, i_con  )

      implicit none

!     input arguments
      integer, intent(in)      :: nx, ny, nz, nw     ! dimensions
      integer, intent(in)      :: i_con
      character(4), intent(in) :: wrfspn         ! wrfchem species' name

!     output arguments
      ! real, intent(out) :: wrfval(nx,ny,nz)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfxs(ny,nz,nw)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfxe(ny,nz,nw)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfys(nx,nz,nw)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfye(nx,nz,nw)         ! wrfchem vmr(ppm)

!     local arguments
      integer :: status, varid
     
      integer :: nspec
      character(20) :: mozspn(20)
      real :: scale

      integer :: i, j, k, n

      print*, 'Process wrfchem species : ', trim(wrfspn)

!     whether there is the species in MOZART data, if not, set 1.0e-6ppm 
      call translate_radm2_mozart( wrfspn, nspec, mozspn, scale )

!     if no this kind of species, set as constant(0.01ppt)
      if( nspec == 0 ) then
        print*, 'Constant value : 1.0e-12ppm' 
        wrfxs(:,:,:) = 1.0e-12
        wrfxe(:,:,:) = 1.0e-12
        wrfys(:,:,:) = 1.0e-12
        wrfye(:,:,:) = 1.0e-12
        return
      endif

!     read MOZART data
      nstt(1:4) = (/ 1, 1, 1, i_con /)
      ncnt(1:4) = (/ nlon, nlat, nlev, 1 /)

      mozval(:,:,:) = 0.0

      do n = 1, nspec

        print*, 'Read and interpolate mozart species : ', trim(mozspn(n))
        status = nf_inq_varid( ncid, mozspn(n), varid )
        if( status /= nf_noerr )  call handle_error_moz( status )

        status = nf_get_vara_real( ncid, varid, nstt(:), ncnt(:), tmpval(:,:,:) )
        if( status /= nf_noerr ) call handle_error_moz( status )

        mozval(:,:,:) = mozval(:,:,:) + tmpval(:,:,:)*1.0e6   ! change unit to ppm

      enddo

      if( scale >= 0.0 ) mozval(:,:,:) = mozval(:,:,:)*scale

!     interpolate in three dimensions

      

      ! west
      i= 1
      do k = 1, nz
         do j = 1, ny
            
            wrfxs(j,k,:)  = mozval(ix(0,i,j),jy(0,i,j),kz(0,i,j,k))*axxs(1,i,j)*byxs(1,i,j)*czxs(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(0,i,j),kz(1,i,j,k))*axxs(1,i,j)*byxs(1,i,j)*czxs(0,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(0,i,j,k))*axxs(1,i,j)*byxs(0,i,j)*czxs(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(1,i,j,k))*axxs(1,i,j)*byxs(0,i,j)*czxs(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(0,i,j,k))*axxs(0,i,j)*byxs(1,i,j)*czxs(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(1,i,j,k))*axxs(0,i,j)*byxs(1,i,j)*czxs(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(0,i,j,k))*axxs(0,i,j)*byxs(0,i,j)*czxs(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(1,i,j,k))*axxs(0,i,j)*byxs(0,i,j)*czxs(0,i,j,k)
            
         enddo
      enddo
      

      ! east
      i = nx
      do k = 1, nz
         do j = 1, ny
            
            wrfxe(j,k,:)  = mozval(ix(0,i,j),jy(0,i,j),kz(0,i,j,k))*axxe(1,i,j)*byxe(1,i,j)*czxe(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(0,i,j),kz(1,i,j,k))*axxe(1,i,j)*byxe(1,i,j)*czxe(0,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(0,i,j,k))*axxe(1,i,j)*byxe(0,i,j)*czxe(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(1,i,j,k))*axxe(1,i,j)*byxe(0,i,j)*czxe(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(0,i,j,k))*axxe(0,i,j)*byxe(1,i,j)*czxe(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(1,i,j,k))*axxe(0,i,j)*byxe(1,i,j)*czxe(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(0,i,j,k))*axxe(0,i,j)*byxe(0,i,j)*czxe(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(1,i,j,k))*axxe(0,i,j)*byxe(0,i,j)*czxe(0,i,j,k)

         enddo
      enddo

      ! north
      j = ny
      do k = 1, nz
         do i = 1, nx

            wrfye(i,k,:)  = mozval(ix(0,i,j),jy(0,i,j),kz(0,i,j,k))*axye(1,i,j)*byye(1,i,j)*czye(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(0,i,j),kz(1,i,j,k))*axye(1,i,j)*byye(1,i,j)*czye(0,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(0,i,j,k))*axye(1,i,j)*byye(0,i,j)*czye(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(1,i,j,k))*axye(1,i,j)*byye(0,i,j)*czye(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(0,i,j,k))*axye(0,i,j)*byye(1,i,j)*czye(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(1,i,j,k))*axye(0,i,j)*byye(1,i,j)*czye(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(0,i,j,k))*axye(0,i,j)*byye(0,i,j)*czye(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(1,i,j,k))*axye(0,i,j)*byye(0,i,j)*czye(0,i,j,k)
            
         enddo
      enddo

      ! south
      j = 1
      do k = 1, nz
         do i = 1, nx

            wrfys(i,k,:)  = mozval(ix(0,i,j),jy(0,i,j),kz(0,i,j,k))*axys(1,i,j)*byys(1,i,j)*czys(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(0,i,j),kz(1,i,j,k))*axys(1,i,j)*byys(1,i,j)*czys(0,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(0,i,j,k))*axys(1,i,j)*byys(0,i,j)*czys(1,i,j,k) + &
                            mozval(ix(0,i,j),jy(1,i,j),kz(1,i,j,k))*axys(1,i,j)*byys(0,i,j)*czys(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(0,i,j,k))*axys(0,i,j)*byys(1,i,j)*czys(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(0,i,j),kz(1,i,j,k))*axys(0,i,j)*byys(1,i,j)*czys(0,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(0,i,j,k))*axys(0,i,j)*byys(0,i,j)*czys(1,i,j,k) + &
                            mozval(ix(1,i,j),jy(1,i,j),kz(1,i,j,k))*axys(0,i,j)*byys(0,i,j)*czys(0,i,j,k)

         enddo
      enddo

      end subroutine mozart_interpolate4d

!---------------------------------------------------------------------------------------------------
!     handle errors produced by calling netCDF functions
!---------------------------------------------------------------------------------------------------

      subroutine handle_error_moz( status )

      implicit none

!     input arguments :
      integer, intent(in) :: status

!     print the error information from processing NETcdf file
      print*, nf_strerror( status )

!     exit from the bconLib
      call exit_mozart_lib( flag=1 )

      end subroutine handle_error_moz

!---------------------------------------------------------------------------------------------------
!     exit from module_mozart_lib 
!---------------------------------------------------------------------------------------------------

      subroutine exit_mozart_lib( flag )

!     input arguments
      integer, optional, intent(in) :: flag

!     local arguments
      integer :: status

!     release memory space
      if( allocated(ix) ) deallocate( ix )
      if( allocated(jy) ) deallocate( jy )
      if( allocated(kz) ) deallocate( kz )
      if( allocated(ax) ) deallocate( ax )
      if( allocated(by) ) deallocate( by )
      if( allocated(cz) ) deallocate( cz )
      if( allocated(axxs) ) deallocate( axxs )
      if( allocated(byxs) ) deallocate( byxs )
      if( allocated(czxs) ) deallocate( czxs )
      if( allocated(axxe) ) deallocate( axxe )
      if( allocated(byxe) ) deallocate( byxe )
      if( allocated(czxe) ) deallocate( czxe )
      if( allocated(axys) ) deallocate( axys )
      if( allocated(byys) ) deallocate( byys )
      if( allocated(czys) ) deallocate( czys )
      if( allocated(axye) ) deallocate( axye )
      if( allocated(byye) ) deallocate( byye )
      if( allocated(czye) ) deallocate( czye )

      if( allocated(mozval) ) deallocate( mozval )
      if( allocated(tmpval) ) deallocate( tmpval )

!     close netCDF file
      if( ncid /= 0 ) status = nf_close( ncid )

!     output information
      if( present(flag) ) then
        select case( flag )
          case( 1 ); print*, 'fail to process netCDF file...'
          case default; print*, 'unknown error(s) occurred ...'
        endselect
        stop ' in moudle_wrfchem_lib ...'
      else
        print*, 'successfully exit from moudle_mozart_lib ...'
      endif

      end subroutine exit_mozart_lib

!---------------------------------------------------------------------------------------------------
!     translate MOZART species names to that recognized by HANK
!---------------------------------------------------------------------------------------------------

      subroutine translate_radm2_mozart( radm2_spn, ns, mozart_spn, scale )

      implicit none

!     input arguments
      character(4), intent(in) :: radm2_spn 

!     output arguments
      integer, intent(out) :: ns
      character(20), intent(out) :: mozart_spn(20)
      real, intent(out) :: scale

!     local arguments
      character(9), parameter :: suffix = '_VMR_avrg'

      ns = 0
      scale = -1.0

      select case( trim(radm2_spn) )


!     Nitrogen species
        case( 'NO' )
          ns = ns + 1; mozart_spn(ns) = 'NO'//suffix

        case( 'NO2' )
          ns = ns + 1; mozart_spn(ns) = 'NO2'//suffix

        case( 'HONO' )

        case( 'NO3' )
          ns = ns + 1; mozart_spn(ns) = 'NO3'//suffix

        case( 'N2O5' )
          ns = ns + 1; mozart_spn(ns) = 'N2O5'//suffix 

        case( 'HNO3' )
          ns = ns + 1; mozart_spn(ns) = 'HNO3'//suffix

        case( 'HNO4' )
          ns = ns + 1; mozart_spn(ns) = 'HO2NO2'//suffix

!     Oxidants
        case( 'O3' )
          ns = ns + 1; mozart_spn(ns) = 'O3'//suffix

        case( 'H2O2' )
          ns = ns + 1; mozart_spn(ns) = 'H2O2'//suffix

!     Sulfur species
        case( 'SO2' )
          ns = ns + 1; mozart_spn(ns) = 'SO2'//suffix

        case( 'SULF' )
          ns = ns + 1; mozart_spn(ns) = 'SO4'//suffix

!     Atomic species
        case( 'O3P' )
          ns = ns + 1; mozart_spn(ns) = 'O'//suffix
          scale = 0.0

        case( 'O1D' )
          ns = ns + 1; mozart_spn(ns) = 'O1D'//suffix
          scale = 0.0

!     Odd hydrogen species
        case( 'HO' )
          ns = ns + 1; mozart_spn(ns) = 'OH'//suffix
          scale = 0.0

        case( 'HO2' )
          ns = ns + 1; mozart_spn(ns) = 'HO2'//suffix

!     Carbon oxides
        case( 'CO' )
          ns = ns + 1; mozart_spn(ns) = 'CO'//suffix

!     Alkanes
        case( 'ETH' )
          ns = ns + 1; mozart_spn(ns) = 'C2H6'//suffix

        case( 'HC3' )
          ns = ns + 1; mozart_spn(ns) = 'C3H8'//suffix
          ns = ns + 1; mozart_spn(ns) = 'C4H10'//suffix

        case( 'HC5' )
          ns = ns + 1; mozart_spn(ns) = 'C3H8'//suffix
          ns = ns + 1; mozart_spn(ns) = 'C4H10'//suffix
!         scale = 12.6

        case( 'HC8' )
          ns = ns + 1; mozart_spn(ns) = 'C3H8'//suffix
          ns = ns + 1; mozart_spn(ns) = 'C4H10'//suffix
!         scale = 2.4 

!     Alkenes
        case( 'OL2' )
          ns = ns + 1; mozart_spn(ns) = 'C2H4'//suffix

        case( 'OLT' )
          ns = ns + 1; mozart_spn(ns) = 'C3H6'//suffix

        case( 'OLI' )
          ns = ns + 1; mozart_spn(ns) = 'C3H6'//suffix
          scale = 0.5

        case( 'ISO' )
          ns = ns + 1; mozart_spn(ns) = 'ISOP'//suffix

!     Aromatics
        case( 'TOL' )

        case( 'XYL' )

        case( 'CSL' )

!     Carbonyls
        case( 'HCHO' )
          ns = ns + 1; mozart_spn(ns) = 'CH2O'//suffix

        case( 'ALD' )
          ns = ns + 1; mozart_spn(ns) = 'CH3CHO'//suffix

        case( 'KET' )
          ns = ns + 1; mozart_spn(ns) = 'CH3COCH3'//suffix

        case( 'GLY' )

        case( 'MGLY' )
          ns = ns + 1; mozart_spn(ns) = 'CH3COCHO'//suffix

        case( 'DCB' )

!     Organic nitrogens
        case( 'PAN' )
          ns = ns + 1; mozart_spn(ns) = 'PAN'//suffix

        case( 'TPAN' )
          ns = ns + 1; mozart_spn(ns) = 'MPAN'//suffix

        case( 'ONIT' )
          ns = ns + 1; mozart_spn(ns) = 'ONIT'//suffix

!     Organic peroxides 
        case( 'OP1' )
          ns = ns + 1; mozart_spn(ns) = 'CH3OOH'//suffix

        case( 'OP2' )
          ns = ns + 1; mozart_spn(ns) = 'POOH'//suffix
          ns = ns + 1; mozart_spn(ns) = 'MACROOH'//suffix
          ns = ns + 1; mozart_spn(ns) = 'C2H5OOH'//suffix
          ns = ns + 1; mozart_spn(ns) = 'C3H7OOH'//suffix
          ns = ns + 1; mozart_spn(ns) = 'ROOH'//suffix
          ns = ns + 1; mozart_spn(ns) = 'XOOH'//suffix
          ns = ns + 1; mozart_spn(ns) = 'ISOPOOH'//suffix

!     Organic acids
        case( 'ORA1' )
          ns = ns + 1; mozart_spn(ns) = 'CO'//suffix
          scale = 6.25e-4

        case( 'ORA2' )
          ns = ns + 1; mozart_spn(ns) = 'CO'//suffix
          scale = 7.5e-4 

!     Aerosol
        case( 'BC1' )
          ns = ns + 1; mozart_spn(ns) = 'CB1'//suffix

        case( 'BC2' )
          ns = ns + 1; mozart_spn(ns) = 'CB2'//suffix

        case default

      end select


      end subroutine translate_radm2_mozart

      end module module_mozart_lib
