
      module module_raqms_lib

      implicit none

!     include files
      include 'netcdf.inc'

!     public readprocedures
      public  :: init_raqms_lib, raqms_interpolate4d, exit_raqms_lib

!     private procedures
      private :: handle_error, translate_radm2_raqms

!     public variables
      integer, public :: nlon, nlat, nlev                    ! data dimension length
      integer, public  :: year_start_r, month_start_r, day_start_r
      integer, public  :: hour_start_r, minute_start_r, second_start_r
      integer, public  :: year_end_r, month_end_r, day_end_r
      integer, public  :: hour_end_r
      integer, public  :: dt_r, ntime_r

!     variables for reading netCDF data
      integer, private :: ncid                               ! netCDF file ID
      integer, private :: nstt(4), ncnt(4)                   ! start index and counter array
      integer, private :: start_index = 1                    ! monthly mean data
      integer, private  :: ntime      ! data dimension length
      integer, private :: nscalar, nstring           ! data dimension length
      integer, private, allocatable :: date(:), datesec(:) 

!     variables used by interpolation
      integer, private, allocatable :: ix(:,:,:), jy(:,:,:), kz(:,:,:,:)  ! index used by interpolation
      real, private, allocatable :: ax(:,:,:), by(:,:,:), cz(:,:,:,:)           ! weight coef. all domain

      real, private, allocatable :: raqval(:,:,:), tmpval(:,:,:), raqval2(:,:,:,:)
      real, allocatable :: pdash(:,:,:), ps_wrf(:,:)

      real, public :: epsilc = 1.E-12

      contains

!---------------------------------------------------------------------------------------------------
!     initialize netcdf file
!---------------------------------------------------------------------------------------------------

      subroutine init_raqms_lib( raqms_fn, ptop_wrf, x2d, y2d, znu, nx, ny, nz )

      implicit none

!     input arguments
      integer, intent(in)      :: nx, ny, nz
      character (len=*), intent(in) :: raqms_fn 
      real, intent(in)         :: ptop_wrf
      real, intent(in)         :: x2d(nx,ny), y2d(nx,ny)
      real, intent(in)         :: znu(nz)

!     local argumets
      integer :: status
      integer :: dimid, varid
      integer :: i, j, k, n 
      integer :: vid
      integer :: dum
      real, allocatable :: x1d(:), y1d(:), x1dp1(:)
!     real, allocatable :: pdash(:,:,:)
      real, allocatable :: p_raq(:), p_wrf(:)
!      real, allocatable :: ps_wrf(:,:)
      real, allocatable :: ps_raq(:,:)
      real x0,x1
      CHARACTER (LEN=100)     :: dtstrings, dtstringe, dtstringf

!     open the input netCDF file
      print *,' INIT_RAQMS: OPEN FILE: ',trim(raqms_fn)
      status = nf_open( trim(raqms_fn), nf_nowrite, ncid )
      if( status /= nf_noerr ) call handle_error( status )

!     get the longitude dimension length of the MOZART data
      status = nf_inq_dimid( ncid, 'lon', dimid )
      if( status /= nf_noerr )  call handle_error( status )

      status = nf_inq_dimlen( ncid, dimid, nlon )
      if( status /= nf_noerr )  call handle_error( status )

!     get the latitude dimension length of the MOZART data
      status = nf_inq_dimid( ncid, 'lat', dimid )
      if( status /= nf_noerr )  call handle_error( status )

      status = nf_inq_dimlen( ncid, dimid, nlat )
      if( status /= nf_noerr )  call handle_error( status )

!     get the vertical dimension length of the MOZART data
      status = nf_inq_dimid( ncid, 'lev', dimid )
      if( status /= nf_noerr )  call handle_error( status )

      status = nf_inq_dimlen( ncid, dimid, nlev )
      if( status /= nf_noerr )  call handle_error( status )

!     read pressure at each grid point
      allocate( pdash(nlon,nlat,nlev) )


      status = nf_inq_varid( ncid, 'pdash', varid )
      if( status /= nf_noerr )  call handle_error( status )

      status = nf_get_var_real( ncid, varid, pdash )
      if( status /= nf_noerr )  call handle_error( status )

!     read longitudes
      allocate( x1d(nlon), x1dp1(nlon+1),y1d(nlat) )

!     print *,' Get lon '
      status = nf_inq_varid( ncid, 'lon', varid )
      if( status /= nf_noerr )  call handle_error( status )

      status = nf_get_var_real( ncid, varid, x1d )
      if( status /= nf_noerr )  call handle_error( status )

!     read latitudes
      status = nf_inq_varid( ncid, 'lat', varid )
      if( status /= nf_noerr )  call handle_error( status )

      status = nf_get_var_real( ncid, varid, y1d )
      if( status /= nf_noerr )  call handle_error( status )

!     read start and end date from files with names
!        like  uwhyb_09_01_2006_00Z.chem.assim.nc

!     print *,' Get time from file name',trim(raqms_fn)

      i = index(raqms_fn,'uwhyb_')
      read(raqms_fn(i+12:i+16),"(i4)") dum
      year_start_r = float(dum)

      i = index(raqms_fn,'uwhyb_')
      read(raqms_fn(i+9 :i+10),"(i2)") dum
      day_start_r = float(dum)

      i = index(raqms_fn,'uwhyb_')
      read(raqms_fn(i+6 :i+ 7),"(i2)") dum
      month_start_r = float(dum)

      i = index(raqms_fn,'uwhyb_')
      read(raqms_fn(i+17:i+18),"(i2)") dum
      hour_start_r = float(dum)

      year_end_r = year_start_r
      month_end_r = month_start_r
      day_end_r = day_start_r + 1.
      hour_end_r = hour_start_r

!    print *, 'raqms ',year_start_r, month_start_r, day_start_r, hour_start_r
!    print *, 'raqms ',year_end_r, month_end_r, day_end_r, hour_end_r

      ! temporal resolution of raqms data files
      dt_r     = 6.            ! hours
      ntime_r  = 1             ! number of data files

!     allocate memory space to store interpolation coef.
      allocate( ax(0:1,nx,ny), by(0:1,nx,ny), cz(0:1,nx,ny,nz) )
      allocate( ix(0:1,nx,ny), jy(0:1,nx,ny), kz(0:1,nx,ny,nz) )

!     horizontal interpolation coefs.

      ! all domain
      print*,'WRF NX,NY ',nx,ny
      print*,'max wrf lon ',maxval(x2d),minval(x2d)
      print*,'max wrf lat ',maxval(y2d),minval(y2d)
      print*,'RAQMS nlon,nlat ',nlon,nlat
      print*,'max raqms lon ',maxval(x1d),minval(x1d)
      print*,'max raqms lat ',maxval(y1d),minval(y1d)


!tks get ax and bx - horizontal interpolants for whole wrf grid
      do i=1,nlon
        x1dp1(i)=x1d(i)
      end do
        x1dp1(nlon+1)=360.


      do i = 1, nx
      do j = 1, ny 
         do n = 1, nlon+1
!           print*,n,x2d(i,j),x1dp1(n)
!           if( x2d(i,j) < x1d(n) ) exit
            if( x2d(i,j) < x1dp1(n) ) exit
         enddo
!        ix(0,i,j) = min( nlon-1, max(n-1,1) )
         ix(0,i,j) = min( nlon, max(n-1,1) )
         ix(1,i,j) = ix(0,i,j) + 1
         x0=x1dp1(ix(0,i,j))
         x1=x1dp1(ix(1,i,j))
         if(ix(1,i,j).eq.nlon+1)then
           ix(1,i,j)=1
           x1=x1dp1(nlon+1)
         endif


!        ax(0,i,j) = ( x2d(i,j) - x1d(ix(0,i,j)) )/( x1d(ix(1,i,j)) - x1d(ix(0,i,j)) )
!        ax(0,i,j) = ( x2d(i,j) - x1dp1(ix(0,i,j)) )/( x1dp1(ix(1,i,j)) - x1dp1(ix(0,i,j)) )
         ax(0,i,j) = ( x2d(i,j) - x0 )/( x1 - x0) 
         ax(1,i,j) = 1.0 - ax(0,i,j)


         do n = 1, nlat
            if( y2d(i,j) < y1d(n) ) exit
         enddo
         
         jy(0,i,j) = min( nlat-1, max(n-1,1) )
         jy(1,i,j) = jy(0,i,j) + 1
         by(0,i,j) = ( y2d(i,j) - y1d(jy(0,i,j)) )/( y1d(jy(1,i,j)) - y1d(jy(0,i,j)) )
         by(1,i,j) = 1.0 - by(0,i,j)
      enddo
      end do

! read surface pressure and interpolate ...

      allocate( ps_raq(nlon,nlat), ps_wrf(nx,ny) )

      status = nf_inq_varid( ncid, 'psfc', varid )
      if( status /= nf_noerr )  call handle_error( status )

      nstt(1:3) = (/ 1, 1, start_index /)
      ncnt(1:3) = (/ nlon, nlat, 1 /)
      status = nf_get_vara_real( ncid, varid, nstt(1:3), ncnt(1:3), ps_raq )
      if( status /= nf_noerr ) call handle_error( status )


      do j = 1, ny 
      do i = 1, nx 
        ps_wrf(i,j) = ps_raq(ix(0,i,j),jy(0,i,j))*ax(1,i,j)*by(1,i,j) +   &
                      ps_raq(ix(0,i,j),jy(1,i,j))*ax(1,i,j)*by(0,i,j) +   &
                      ps_raq(ix(1,i,j),jy(0,i,j))*ax(0,i,j)*by(1,i,j) +   &
                      ps_raq(ix(1,i,j),jy(1,i,j))*ax(0,i,j)*by(0,i,j)
      enddo
      enddo

!     release memory space not used
!tks     deallocate( x1d, y1d, pdash, ps_raq, ps_wrf, p_raq, p_wrf )
      deallocate( x1d, y1d,x1dp1,  ps_raq )

!     allocate memory space for reading
      allocate( raqval(nlon,nlat,nlev), tmpval(nlon,nlat,nlev), raqval2(nlon,nlat,nlev,2) )

      end subroutine init_raqms_lib
!---------------------------------------------------------------------------------------------------

      subroutine re_init_raqms( raqms_fn, it, idt )

      implicit none

!     input arguments
      character (len=*), intent(inout) :: raqms_fn 
      integer dymnth(12)                               !tks
      integer, intent(in)         :: it, idt

!     local argumets
      character (len=2)  :: cday, chour, cfxhour, cmon
      integer :: status
      integer :: dimid, varid
      integer :: i, j, k, n 
      integer :: dum
      data dymnth/31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31/    !tks

!     close netCDF file
      if( ncid /= 0 ) status = nf_close( ncid )

!     read start and end date from files with names
!        like  uwhyb_09_01_2006_00Z.chem.assim.nc
      print *,' SUBROUTINE RE_INIT: CLOSED FILE: ',trim(raqms_fn)

      i = index(raqms_fn,'uwhyb_')
      read(raqms_fn(i+12:i+16),"(i4)") dum
      year_start_r = float(dum)

      read(raqms_fn(i+9 :i+10),"(i2)") dum
      day_start_r = float(dum)

      read(raqms_fn(i+6 :i+ 7),"(i2)") dum
      month_start_r = float(dum)

      read(raqms_fn(i+17:i+18),"(i2)") dum
      hour_start_r = float(dum)

      hour_start_r = hour_start_r + float(idt)

      if( hour_start_r >= 24. ) then
         day_start_r = day_start_r + 1
         if(day_start_r .gt. dymnth(month_start_r))then
             month_start_r=month_start_r+1
             day_start_r=1
         endif
          
         hour_start_r = hour_start_r  - 24.

      endif

      if( month_start_r < 10 ) then
        write(cmon,"(a1,i1)") "0",int(month_start_r)
      else
        write(cmon,"(i2)") int(month_start_r)
      endif

      if( day_start_r < 10 ) then
        write(cday,"(a1,i1)") "0",int(day_start_r)
      else
        write(cday,"(i2)") int(day_start_r)
      endif
      if( hour_start_r < 10 ) then
        write(chour,"(a1,i1)") "0",int(hour_start_r)
      else
        write(chour,"(i2)") int(hour_start_r)
      endif


      write(raqms_fn(i+17:i+18) ,"(a2)") chour
      write(raqms_fn(i+9 :i+10) ,"(a2)") cday
      write(raqms_fn(i+6 :i+7) ,"(a2)") cmon


!     open the input netCDF file
      print *,' SUBROUTINE RE_INIT_RAQMS: OPEN ED FILE: ',trim(raqms_fn)
      status = nf_open( trim(raqms_fn), nf_nowrite, ncid )
      if( status /= nf_noerr ) call handle_error( status )

      end subroutine re_init_raqms
!---------------------------------------------------------------------------------------------------
!     interpolate four-dimensional field ...
!---------------------------------------------------------------------------------------------------

      subroutine raqms_interpolate4d_ic( wrfspn, wrfgrid, nx, ny, nz, ptop_wrf, znu  )

      implicit none

!     input arguments
      integer, intent(in)      :: nx, ny, nz     ! dimensions
!tks     character(LEN=6), intent(in) :: wrfspn         ! wrfchem species' name
      character(LEN=12), intent(in) :: wrfspn         ! wrfchem species' name


!     output arguments
!     output arguments
!     output arguments
      real, intent(out) :: wrfgrid(nx,ny,nz)         ! wrfchem vmr(ppm)

      real raq_on_wrf_horz(nx,ny,nlev),praq_on_wrf_horz(nx,ny,nlev),xxx(nx,ny),yyy(nx,ny)
      real p_raq(nlev),p_wrf(nz),znu(nz),ptop_wrf

      real mw_air,mw_oc,mw_bc,mw_sulf,rat,scal

!     local arguments
      integer :: status, varid

      integer :: nspec
      character(LEN=20) :: raqspn(20)
      real :: scale

      integer :: i, j, k, n

        print*, 'Process wrfchem species : ', trim(wrfspn)


!     whether there is the species in MOZART data, if not, set 1.0e-6ppm

      call translate_radm2_raqms( wrfspn, nspec, raqspn, scale )


!     read MOZART data
!     nstt(1:4) = (/ 1, 1, 1, i_con /)
      nstt(1:4) = (/ 1, 1, 1, 1 /)
      ncnt(1:4) = (/ nlon, nlat, nlev, 1 /)

      raqval(:,:,:) = 0.0

      do n = 1, nspec

        print*, 'Read and interpolate raqms species : ', n,trim(raqspn(n))
        status = nf_inq_varid( ncid, raqspn(n), varid )
        if( status /= nf_noerr )  call handle_error( status )

        status = nf_get_vara_real( ncid, varid, nstt(:), ncnt(:), tmpval(:,:,:) )
        if( status /= nf_noerr ) call handle_error( status )
!
!       write(6,*)'varid ',varid,maxval(tmpval),minval(tmpval)

!rbp flip raqval from top-to-bottom to bottom-to-top to be consistent with
!rbp p_val and vertical interpolaters
!convert RAQMS units to WRF/Chem units
!gases need to be converted to ppmv (come in ppm)
!aerosols to micro grams/kg

       mw_air=28.97
       mw_bc=12.
       mw_oc=16.8
       mw_sulf=132.

        scal=1.e6
        rat=1.

        if(wrfspn(1:3).eq.'BC1'.or.wrfspn(1:3).eq.'BC2')then
           rat=mw_bc/mw_air
           scal=1.e9
        endif

        if(wrfspn(1:3).eq.'OC1'.or.wrfspn(1:3).eq.'OC2')then
           rat=mw_oc/mw_air
           scal=1.e9
        endif

! sulphate is in volume mixing ratio
!       if(wrfspn(1:4).eq.'sulf')then
!          rat=mw_sulf/mw_air
!          scal=1.e6
!       endif

        if(trim(wrfspn(1:4)).eq.'SEAS')scal=1.e9
        if(trim(wrfspn(1:4)).eq.'DUST')scal=1.e9
         do k = 1, nlev
          raqval(:,:,k) = raqval(:,:,k) + tmpval(:,:,nlev-k+1)*scal*rat   ! change unit to ppm,  rbp
         enddo
      enddo
      if( scale >= 0.0 ) raqval(:,:,:) = raqval(:,:,:)*scale

      print*,'raqval ',maxval(raqval),minval(raqval)

!tks  interpolate pressure and chemical to WRF horizontal grid
       praq_on_wrf_horz(:,:,:)=0.
       raq_on_wrf_horz(:,:,:)=0.
       do k=1,nlev
       do i=1,nx
       do j=1,ny
         raq_on_wrf_horz(i,j,k) =raqval(ix(0,i,j),jy(0,i,j),k)*ax(1,i,j)*by(1,i,j) + &
                                 raqval(ix(0,i,j),jy(1,i,j),k)*ax(1,i,j)*by(0,i,j) + &
                                 raqval(ix(1,i,j),jy(0,i,j),k)*ax(0,i,j)*by(1,i,j) + &
                                 raqval(ix(1,i,j),jy(1,i,j),k)*ax(0,i,j)*by(0,i,j)

         praq_on_wrf_horz(i,j,k)=pdash(ix(0,i,j),jy(0,i,j),nlev-k+1)*ax(1,i,j)*by(1,i,j) + &
                                 pdash(ix(0,i,j),jy(1,i,j),nlev-k+1)*ax(1,i,j)*by(0,i,j) + &
                                 pdash(ix(1,i,j),jy(0,i,j),nlev-k+1)*ax(0,i,j)*by(1,i,j) + &
                                 pdash(ix(1,i,j),jy(1,i,j),nlev-k+1)*ax(0,i,j)*by(0,i,j)
       end do
       end do
       end do
      print*,'tks ax,by ',maxval(ax),minval(ax),maxval(by),minval(by)


!tks now get vertical interpolator
!tks data and pressure are on WRF horizontal grid, now interpolate in vertical
      cz(:,:,:,:)=0.
      do i =1, nx
      do j = 1, ny
         do k = 1, nlev
           p_raq(k) = log(praq_on_wrf_horz(i,j,k))
         enddo

         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

         do k = 1, nz
            do n = 1, nlev
               if( p_wrf(k) > p_raq(n) ) exit
            enddo
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            cz(0,i,j,k) = ( p_wrf(k) - p_raq(kz(0,i,j,k)) )/( p_raq(kz(1,i,j,k)) - p_raq(kz(0,i,j,k)) )
            cz(1,i,j,k) = 1.0 - cz(0,i,j,k)
         enddo
      enddo
      enddo
      print*,'tks cz ',maxval(cz),minval(cz)
!tks interpolate in vertical
       do k=1,nz
       do i=1,nx
       do j=1,ny
          wrfgrid(i,j,k)= raq_on_wrf_horz(i,j,kz(0,i,j,k))*cz(1,i,j,k)+raq_on_wrf_horz(i,j,kz(1,i,j,k))*cz(0,i,j,k)
       enddo
       enddo
       enddo
       print*,'wrfval ',maxval(wrfgrid),minval(wrfgrid)


!!!!tks      deallocate(pdash)
      end subroutine raqms_interpolate4d_ic

!---------------------------------------------------------------------------------------------------
!     interpolate four-dimensional field ... 
!---------------------------------------------------------------------------------------------------

      subroutine raqms_interpolate4d( wrfspn, wrfxs, wrfxe, wrfys, wrfye, nx, ny, nz, nw, i_con, &
                                      znu,ptop_wrf  )

      implicit none

!     input arguments
      integer, intent(in)      :: nx, ny, nz, nw     ! dimensions
      integer, intent(in)      :: i_con
!tks     character(LEN=6), intent(in) :: wrfspn         ! wrfchem species' name
      character(LEN=12), intent(in) :: wrfspn         ! wrfchem species' name

!     output arguments
    ! real, intent(out) :: wrfval(nx,ny,nz)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfxs(ny,nz,nw)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfxe(ny,nz,nw)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfys(nx,nz,nw)         ! wrfchem vmr(ppm)
      real, intent(out) :: wrfye(nx,nz,nw)         ! wrfchem vmr(ppm)

      real raq_on_wrf_horz(nx,ny,nlev),praq_on_wrf_horz(nx,ny,nlev),xxx(nx,ny),yyy(nx,ny)
      real wrfgrid(nx,ny,nz),p_raq(nlev),p_wrf(nz),znu(nz),ptop_wrf

      real mw_air,mw_oc,mw_bc,mw_sulf,rat,scal

!     local arguments
      integer :: status, varid
     
      integer :: nspec
      character(LEN=20) :: raqspn(20)
      real :: scale

      integer :: i, j, k, n

!       print*, 'Process wrfchem species : ', trim(wrfspn)


!     whether there is the species in MOZART data, if not, set 1.0e-6ppm 

      call translate_radm2_raqms( wrfspn, nspec, raqspn, scale )


!     if no this kind of species, set as constant(0.01ppt)
      if( nspec == 0 ) then
        print*, 'Constant value : ' , epsilc
        wrfxs(:,:,:) = epsilc
        wrfxe(:,:,:) = epsilc 
        wrfys(:,:,:) = epsilc 
        wrfye(:,:,:) = epsilc 
        return
      endif

!     read MOZART data
!     nstt(1:4) = (/ 1, 1, 1, i_con /)
      nstt(1:4) = (/ 1, 1, 1, 1 /)
      ncnt(1:4) = (/ nlon, nlat, nlev, 1 /)

      raqval(:,:,:) = 0.0

      do n = 1, nspec

!       print*, 'Read and interpolate raqms species : ', trim(raqspn(n))
        status = nf_inq_varid( ncid, raqspn(n), varid )
        if( status /= nf_noerr )  call handle_error( status )

        status = nf_get_vara_real( ncid, varid, nstt(:), ncnt(:), tmpval(:,:,:) )
        if( status /= nf_noerr ) call handle_error( status )
!
!       write(6,*)'varid ',varid,maxval(tmpval),minval(tmpval)

!rbp flip raqval from top-to-bottom to bottom-to-top to be consistent with
!rbp p_val and vertical interpolaters
!convert RAQMS units to WRF/Chem units
!gases need to be converted to ppmv (come in ppm)
!aerosols to micro grams/kg

       mw_air=28.97
       mw_bc=12.
       mw_oc=16.8
       mw_sulf=132.

        scal=1.e6
        rat=1.

        if(wrfspn(1:3).eq.'BC1'.or.wrfspn(1:3).eq.'BC2')then
           rat=mw_bc/mw_air
           scal=1.e9
        endif

        if(wrfspn(1:3).eq.'OC1'.or.wrfspn(1:3).eq.'OC2')then
           rat=mw_oc/mw_air
           scal=1.e9
        endif

! sulphate is in volume mixing ratio
!       if(wrfspn(1:4).eq.'sulf')then
!          rat=mw_sulf/mw_air
!          scal=1.e6
!       endif

        if(trim(wrfspn(1:4)).eq.'SEAS')scal=1.e9
        if(trim(wrfspn(1:4)).eq.'DUST')scal=1.e9


         do k = 1, nlev
          raqval(:,:,k) = raqval(:,:,k) + tmpval(:,:,nlev-k+1)*scal*rat   ! change unit to ppm,  rbp
         enddo
      enddo
      if( scale >= 0.0 ) raqval(:,:,:) = raqval(:,:,:)*scale


       praq_on_wrf_horz(:,:,:)=0.
       raq_on_wrf_horz(:,:,:)=0.



!tks  interpolate pressure and chemical to WRF horizontal grid

!North edge
       do k=1,nlev
       do i=1,nx
       do j=ny,ny
         raq_on_wrf_horz(i,j,k) =raqval(ix(0,i,j),jy(0,i,j),k)*ax(1,i,j)*by(1,i,j) + &
                                 raqval(ix(0,i,j),jy(1,i,j),k)*ax(1,i,j)*by(0,i,j) + &
                                 raqval(ix(1,i,j),jy(0,i,j),k)*ax(0,i,j)*by(1,i,j) + &
                                 raqval(ix(1,i,j),jy(1,i,j),k)*ax(0,i,j)*by(0,i,j) 

         praq_on_wrf_horz(i,j,k)=pdash(ix(0,i,j),jy(0,i,j),nlev-k+1)*ax(1,i,j)*by(1,i,j) + &
                                 pdash(ix(0,i,j),jy(1,i,j),nlev-k+1)*ax(1,i,j)*by(0,i,j) + &
                                 pdash(ix(1,i,j),jy(0,i,j),nlev-k+1)*ax(0,i,j)*by(1,i,j) + &
                                 pdash(ix(1,i,j),jy(1,i,j),nlev-k+1)*ax(0,i,j)*by(0,i,j) 
       end do
       end do
       end do
!tks now get vertical interpolator
!tks data and pressure are on WRF horizontal grid, now interpolate in vertical
      cz(:,:,:,:)=0.
      do i =1, nx
      do j = ny,ny
         do k = 1, nlev
           p_raq(k) = log(praq_on_wrf_horz(i,j,k))
         enddo

         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

         do k = 1, nz
            do n = 1, nlev
               if( p_wrf(k) > p_raq(n) ) exit
            enddo
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            cz(0,i,j,k) = ( p_wrf(k) - p_raq(kz(0,i,j,k)) )/( p_raq(kz(1,i,j,k)) - p_raq(kz(0,i,j,k)) )
            cz(1,i,j,k) = 1.0 - cz(0,i,j,k)
         enddo
      enddo
      enddo
!tks interpolate in vertical
       do k=1,nz
       do i=1,nx
       do j=ny,ny
          wrfgrid(i,j,k)= raq_on_wrf_horz(i,j,kz(0,i,j,k))*cz(1,i,j,k)+raq_on_wrf_horz(i,j,kz(1,i,j,k))*cz(0,i,j,k)
       enddo
       enddo
       enddo
! north
      j = ny
      do k = 1, nz
         do i = 1, nx
            wrfye(i,k,:)  = wrfgrid(i,j,k)
            wrfye(i,k,:)  = max( wrfye(i,k,:), 1.E-12 )
         enddo
      enddo

!South edge
       do k=1,nlev
       do i=1,nx
       do j=1,1
         raq_on_wrf_horz(i,j,k) =raqval(ix(0,i,j),jy(0,i,j),k)*ax(1,i,j)*by(1,i,j) + &
                                 raqval(ix(0,i,j),jy(1,i,j),k)*ax(1,i,j)*by(0,i,j) + &
                                 raqval(ix(1,i,j),jy(0,i,j),k)*ax(0,i,j)*by(1,i,j) + &
                                 raqval(ix(1,i,j),jy(1,i,j),k)*ax(0,i,j)*by(0,i,j)

         praq_on_wrf_horz(i,j,k)=pdash(ix(0,i,j),jy(0,i,j),nlev-k+1)*ax(1,i,j)*by(1,i,j) + &
                                 pdash(ix(0,i,j),jy(1,i,j),nlev-k+1)*ax(1,i,j)*by(0,i,j) + &
                                 pdash(ix(1,i,j),jy(0,i,j),nlev-k+1)*ax(0,i,j)*by(1,i,j) + &
                                 pdash(ix(1,i,j),jy(1,i,j),nlev-k+1)*ax(0,i,j)*by(0,i,j)
       end do
       end do
       end do
!tks now get vertical interpolator
!tks data and pressure are on WRF horizontal grid, now interpolate in vertical
      cz(:,:,:,:)=0.
      do i =1, nx
      do j = 1,1
         do k = 1, nlev
           p_raq(k) = log(praq_on_wrf_horz(i,j,k))
         enddo

         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

         do k = 1, nz
            do n = 1, nlev
               if( p_wrf(k) > p_raq(n) ) exit
            enddo
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            cz(0,i,j,k) = ( p_wrf(k) - p_raq(kz(0,i,j,k)) )/( p_raq(kz(1,i,j,k)) - p_raq(kz(0,i,j,k)) )
            cz(1,i,j,k) = 1.0 - cz(0,i,j,k)
         enddo
      enddo
      enddo
!tks interpolate in vertical
       do k=1,nz
       do i=1,nx
       do j=1,1
          wrfgrid(i,j,k)= raq_on_wrf_horz(i,j,kz(0,i,j,k))*cz(1,i,j,k)+raq_on_wrf_horz(i,j,kz(1,i,j,k))*cz(0,i,j,k)
       enddo
       enddo
       enddo
! south
      j = 1
      do k = 1, nz
         do i = 1, nx
            wrfys(i,k,:)  = wrfgrid(i,j,k)
            wrfys(i,k,:)  = max( wrfys(i,k,:), 1.E-12 )
         enddo
      enddo


!West edge
       do k=1,nlev
       do i=1,1
       do j=1,ny
         raq_on_wrf_horz(i,j,k) =raqval(ix(0,i,j),jy(0,i,j),k)*ax(1,i,j)*by(1,i,j) + &
                                 raqval(ix(0,i,j),jy(1,i,j),k)*ax(1,i,j)*by(0,i,j) + &
                                 raqval(ix(1,i,j),jy(0,i,j),k)*ax(0,i,j)*by(1,i,j) + &
                                 raqval(ix(1,i,j),jy(1,i,j),k)*ax(0,i,j)*by(0,i,j)

         praq_on_wrf_horz(i,j,k)=pdash(ix(0,i,j),jy(0,i,j),nlev-k+1)*ax(1,i,j)*by(1,i,j) + &
                                 pdash(ix(0,i,j),jy(1,i,j),nlev-k+1)*ax(1,i,j)*by(0,i,j) + &
                                 pdash(ix(1,i,j),jy(0,i,j),nlev-k+1)*ax(0,i,j)*by(1,i,j) + &
                                 pdash(ix(1,i,j),jy(1,i,j),nlev-k+1)*ax(0,i,j)*by(0,i,j)
       end do
       end do
       end do
!tks now get vertical interpolator
!tks data and pressure are on WRF horizontal grid, now interpolate in vertical
      cz(:,:,:,:)=0.
      do i =1, 1
      do j = 1,ny
         do k = 1, nlev
           p_raq(k) = log(praq_on_wrf_horz(i,j,k))
         enddo

         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

         do k = 1, nz
            do n = 1, nlev
               if( p_wrf(k) > p_raq(n) ) exit
            enddo
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            cz(0,i,j,k) = ( p_wrf(k) - p_raq(kz(0,i,j,k)) )/( p_raq(kz(1,i,j,k)) - p_raq(kz(0,i,j,k)) )
            cz(1,i,j,k) = 1.0 - cz(0,i,j,k)
         enddo
      enddo
      enddo
!tks interpolate in vertical
       do k=1,nz
       do i=1,1
       do j=1,ny
          wrfgrid(i,j,k)= raq_on_wrf_horz(i,j,kz(0,i,j,k))*cz(1,i,j,k)+raq_on_wrf_horz(i,j,kz(1,i,j,k))*cz(0,i,j,k)
       enddo
       enddo
       enddo
      i= 1
      do k = 1, nz
         do j = 1, ny
            wrfxs(j,k,:)  = wrfgrid(i,j,k)
            wrfxs(j,k,:)  = max( wrfxs(j,k,:), 1.E-12 )
         enddo
      enddo


!East edge
       do k=1,nlev
       do i=nx,nx
       do j=1,ny
         raq_on_wrf_horz(i,j,k) =raqval(ix(0,i,j),jy(0,i,j),k)*ax(1,i,j)*by(1,i,j) + &
                                 raqval(ix(0,i,j),jy(1,i,j),k)*ax(1,i,j)*by(0,i,j) + &
                                 raqval(ix(1,i,j),jy(0,i,j),k)*ax(0,i,j)*by(1,i,j) + &
                                 raqval(ix(1,i,j),jy(1,i,j),k)*ax(0,i,j)*by(0,i,j)

         praq_on_wrf_horz(i,j,k)=pdash(ix(0,i,j),jy(0,i,j),nlev-k+1)*ax(1,i,j)*by(1,i,j) + &
                                 pdash(ix(0,i,j),jy(1,i,j),nlev-k+1)*ax(1,i,j)*by(0,i,j) + &
                                 pdash(ix(1,i,j),jy(0,i,j),nlev-k+1)*ax(0,i,j)*by(1,i,j) + &
                                 pdash(ix(1,i,j),jy(1,i,j),nlev-k+1)*ax(0,i,j)*by(0,i,j)
       end do
       end do
       end do
!tks now get vertical interpolator
!tks data and pressure are on WRF horizontal grid, now interpolate in vertical
      cz(:,:,:,:)=0.
      do i =nx, nx
      do j = 1,ny
         do k = 1, nlev
           p_raq(k) = log(praq_on_wrf_horz(i,j,k))
         enddo

         p_wrf(:) = log( znu(:)*ps_wrf(i,j) + (1.0-znu(:))*ptop_wrf )

         do k = 1, nz
            do n = 1, nlev
               if( p_wrf(k) > p_raq(n) ) exit
            enddo
            kz(0,i,j,k) = min( nlev-1, max(n-1,1) )
            kz(1,i,j,k) = kz(0,i,j,k) + 1
            cz(0,i,j,k) = ( p_wrf(k) - p_raq(kz(0,i,j,k)) )/( p_raq(kz(1,i,j,k)) - p_raq(kz(0,i,j,k)) )
            cz(1,i,j,k) = 1.0 - cz(0,i,j,k)
         enddo
      enddo
      enddo
!tks interpolate in vertical
       do k=1,nz
       do i=nx,nx
       do j=1,ny
          wrfgrid(i,j,k)= raq_on_wrf_horz(i,j,kz(0,i,j,k))*cz(1,i,j,k)+raq_on_wrf_horz(i,j,kz(1,i,j,k))*cz(0,i,j,k)
       enddo
       enddo
       enddo
      i= nx
      do k = 1, nz
         do j = 1, ny
            wrfxe(j,k,:)  = wrfgrid(i,j,k)
            wrfxe(j,k,:)  = max( wrfxe(j,k,:), 1.E-12 )
         enddo
      enddo


!!!!tks      deallocate(pdash)
      end subroutine raqms_interpolate4d

!---------------------------------------------------------------------------------------------------
!     handle errors produced by calling netCDF functions
!---------------------------------------------------------------------------------------------------

      subroutine handle_error( status )

      implicit none

!     input arguments :
      integer, intent(in) :: status

!     print the error information from processing NETcdf file
      print*, nf_strerror( status )

!     exit from the bconLib
      call exit_raqms_lib( flag=1 )

      end subroutine handle_error

!---------------------------------------------------------------------------------------------------
!     exit from module_raqms_lib 
!---------------------------------------------------------------------------------------------------

      subroutine exit_raqms_lib( flag )

!     input arguments
      integer, optional, intent(in) :: flag

!     local arguments
      integer :: status

!     release memory space
      if( allocated(ix) ) deallocate( ix )
      if( allocated(jy) ) deallocate( jy )
      if( allocated(kz) ) deallocate( kz )
      if( allocated(ax) ) deallocate( ax )
      if( allocated(by) ) deallocate( by )
      if( allocated(cz) ) deallocate( cz )
      if( allocated(ps_wrf) ) deallocate( ps_wrf)

      if( allocated(raqval) ) deallocate( raqval )
      if( allocated(tmpval) ) deallocate( tmpval )

!     close netCDF file
      if( ncid /= 0 ) status = nf_close( ncid )

!     output information
      if( present(flag) ) then
        select case( flag )
          case( 1 ); print*, 'module_raqms_lib: fail to process netCDF file...'
          case default; print*, 'unknown error(s) occurred ...'
        endselect
        stop ' in module_wrfchem_lib ...'
      else
        print*, 'successfully exit from module_raqms_lib ...'
      endif

      end subroutine exit_raqms_lib

!---------------------------------------------------------------------------------------------------
!     translate RAQMS species names to that recognized by WRF/CHEM
!---------------------------------------------------------------------------------------------------

      subroutine translate_radm2_raqms( radm2_spn, ns, raqms_spn, scale )

      implicit none

!     input arguments
!     character(LEN=6), intent(in) :: radm2_spn 
      character(LEN=12), intent(in) :: radm2_spn 

!     output arguments
      integer, intent(out) :: ns
      character(LEN=20), intent(out) :: raqms_spn(20)
      real, intent(out) :: scale


!     local arguments
      character(LEN=9), parameter :: suffix = ''

      ns = 0
      scale = -1.0



      select case( trim(radm2_spn) )


!tks 11/11/08   added for racm mechanism
        case( 'ete' )
          ns = ns + 1; raqms_spn(ns) = 'ieth'//suffix
        case( 'ch4' )
          ns = ns + 1; raqms_spn(ns) = 'ich4'//suffix
        case( 'macr' )
          ns = ns + 1; raqms_spn(ns) = 'imacr'//suffix
!tks 11/11/08



!     Nitrogen species
        case( 'no' )
          ns = ns + 1; raqms_spn(ns) = 'no'//suffix

        case( 'no2' )
          ns = ns + 1; raqms_spn(ns) = 'ino2'//suffix

        case( 'hono' )

        case( 'no3' )
          ns = ns + 1; raqms_spn(ns) = 'ino3'//suffix

        case( 'n2o5' )
          ns = ns + 1; raqms_spn(ns) = 'in2o5'//suffix 

        case( 'hno3' )
          ns = ns + 1; raqms_spn(ns) = 'ihno3t'//suffix

        case( 'hno4' )
          ns = ns + 1; raqms_spn(ns) = 'ihno4'//suffix

!     Oxidants
        case( 'o3' )
          ns = ns + 1; raqms_spn(ns) = 'o3vmr'//suffix
          scale = 1.0

        case( 'h2o2' )
          ns = ns + 1; raqms_spn(ns) = 'ih2o2'//suffix

!     Sulfur species
        case( 'so2' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'iso2'//suffix
           

        case( 'sulf' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'iso4aer'//suffix

!     Atomic species
        case( 'O3P' )

        case( 'O1D' )

!     Odd hydrogen species
        case( 'ho' )
          ns = ns + 1; raqms_spn(ns) = 'oh'//suffix

        case( 'ho2' )
          ns = ns + 1; raqms_spn(ns) = 'ho2'//suffix

!     Carbon oxides
        case( 'co' )
          ns = ns + 1; raqms_spn(ns) = 'ico'//suffix

!     Alkanes
        case( 'eth' )
          ns = ns + 1; raqms_spn(ns) = 'ic2h6'//suffix

        case( 'HC3' )

        case( 'HC5' )

        case( 'HC8' )

!     Alkenes
        case( 'OL2' )

        case( 'olt' )
          ns = ns + 1; raqms_spn(ns) = 'iolet'//suffix

        case( 'oli' )
          ns = ns + 1; raqms_spn(ns) = 'iolei'//suffix

        case( 'iso' )
          ns = ns + 1; raqms_spn(ns) = 'iisop'//suffix

!     Aromatics
        case( 'TOL' )

        case( 'XYL' )

        case( 'CSL' )

!     Carbonyls
        case( 'hcho' )
          ns = ns + 1; raqms_spn(ns) = 'ich2o'//suffix

        case( 'ald' )
          ns = ns + 1; raqms_spn(ns) = 'iald2'//suffix

        case( 'KET' )

        case( 'GLY' )

        case( 'mgly' )
          ns = ns + 1; raqms_spn(ns) = 'imgly'//suffix

        case( 'DCB' )

!     Organic nitrogens
        case( 'pan' )
          ns = ns + 1; raqms_spn(ns) = 'ipan'//suffix

        case( 'TPAN' )

        case( 'onit' )
          ns = ns + 1; raqms_spn(ns) = 'ionit'//suffix

!     Organic peroxides 
        case( 'op1' )
!         ns = ns + 1; raqms_spn(ns) = 'ich3ooh'//suffix

        case( 'op2' )
!         ns = ns + 1; raqms_spn(ns) = 'irooh'//suffix

!     Organic acids
        case( 'ORA1' )

        case( 'ORA2' )

!     Aerosol
        case( 'BC1' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'ibc1'//suffix

        case( 'BC2' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'ibc2'//suffix

        case( 'OC1' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'ioc1'//suffix

        case( 'OC2' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'ioc2'//suffix

        case( 'dms' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'idms'//suffix

        case( 'msa' )
!tks
          ns = ns + 1; raqms_spn(ns) = 'imsa'//suffix

!tks
        case( 'DUST_1' )
          ns = ns + 1; raqms_spn(ns) = 'idu1'//suffix
!tks
        case( 'DUST_2' )
          ns = ns + 1; raqms_spn(ns) = 'idu2'//suffix
!tks
        case( 'DUST_3' )
          ns = ns + 1; raqms_spn(ns) = 'idu3'//suffix
!tks
        case( 'DUST_4' )
          ns = ns + 1; raqms_spn(ns) = 'idu4'//suffix
!tks
        case( 'DUST_5' )
          ns = ns + 1; raqms_spn(ns) = 'idu5'//suffix
!tks
        case( 'SEAS_1' )
          ns = ns + 1; raqms_spn(ns) = 'iss1'//suffix
!tks
        case( 'SEAS_2' )
          ns = ns + 1; raqms_spn(ns) = 'iss2'//suffix
!tks
        case( 'SEAS_3' )
          ns = ns + 1; raqms_spn(ns) = 'iss3'//suffix
!tks
        case( 'SEAS_4' )
          ns = ns + 1; raqms_spn(ns) = 'iss4'//suffix

        case default

      end select


      end subroutine translate_radm2_raqms

      end module module_raqms_lib
